/* $Id$ */ 
/**
 * @file rrqrGate.c
 * @
 * @date 01/11/12
 * @author  Martin Koehler (), koehlerm@mpi-magdeburg.mpg.de - (MPI Magdeburg, CSC Group)
 * @version $Id$
 * 
 *    Description:  <CURSOR>
 *
 */

#ifndef F77_INT
#define F77_INT int
#endif 

#include "mex.h"
#include <complex.h>
#include <stdio.h> 
#include <string.h> 
#define min(A,B) (((A)<(B))?(A):(B))
#define max(A,B) (((A)>(B))?(A):(B))


void zgeqpx_(F77_INT *JOB, F77_INT *M, F77_INT *N, F77_INT *K, double complex *Ac, F77_INT *LDA, 
	     double complex *Cc, F77_INT *LDC, F77_INT *JPVT, double *IRCOND, double *ORCONDr, 
	     F77_INT *RANK, double * SVLUESr, double complex * WORKc, F77_INT *LWORK, double *WORK, F77_INT *INFO); 
void dgeqpx_(F77_INT *JOB, F77_INT *M, F77_INT *N, F77_INT *K, double *A, F77_INT *LDA, 
		double *C, F77_INT *LDC, F77_INT *JPVT, double *IRCOND, double *ORCONDr, 
		F77_INT *RANK, double *SVLUESr,double *WORK, F77_INT *LWORK, F77_INT *INFO); 

void mexFunction(
	int nlhs, mxArray *plhs[],
	int nrhs, const mxArray *prhs[])
{
	F77_INT JOB, M,N,K,LDA,LDC,LWORK,RANK,INFO,MN; 
	int MC,NC, mA,nA,i,j,RW; 
	F77_INT *JPVT; 
	double IRCOND, ORCONDr, SVLUESr[4],helper; 
	double *JPVTr; 
	double *Ar,*Cr,*WORKr; 
	double complex *Ac, *Cc, *WORKc; 
	double *RWORK; 
	int ic;
#ifdef DEBUG
	printf("NLHS: %d\nNRHS: %d\n",nlhs,nrhs); 
#endif

	/*   1st case: [R,p,r]=rrqr(A,tol) */ 
	if(nlhs == 3) {
		if ( nrhs == 3) {
			mexErrMsgTxt("One or two input arguments required."); 
		} else if ( nrhs == 2 ) {
			double *tmp = mxGetPr(prhs[1]); 
			IRCOND = *tmp; 
		        if(IRCOND < 0 ){ 
				       mexErrMsgTxt("tol must be greater then zero."); 
		        }
		} else {
			IRCOND = 0.0; 	/*  maps to machine precision internally */
		}
#ifdef DEBUG
		printf("IRCOND = %lg\n", IRCOND); 
#endif
		JOB = 1; 
		M = mxGetM(prhs[0]); 
		N = mxGetN(prhs[0]); 
		K = 0; 
		LDA = M; 
		LDC = 1; 
		MN = min( M, N ); 
		LWORK = 2*MN+3*N; 
		JPVT = (F77_INT *) malloc(sizeof(F77_INT)*N); 

            	ic = mxIsComplex(prhs[0]); 
#ifdef DEBUG
		printf("JOB(1)=%d\n",JOB); 
		printf("M  (2)=%d\n",M); 
		printf("N  (3)=%d\n",N); 
		printf("K  (4)=%d\n",K); 
		printf("LDA(6)=%d\n",LDA); 
#endif


	        if (ic ){
			int i;
			double *RE; 
			double *IM; 
			Ac = (double complex *) malloc(sizeof(double complex) * LDA *N); 
			Cc = (double complex *) malloc(sizeof(double complex) *LDA *N); 
			WORKc = (double complex *) malloc( sizeof (double complex) *LWORK); 
            		RW = 2*N; 
			RWORK = (double *) malloc(sizeof(double)* RW); 


			RE = mxGetPr(prhs[0]); 
			IM = mxGetPi(prhs[0]); 
			for ( i = 0; i < LDA*N; i++) {
				Ac[i] = RE[i] + IM[i]*I; 
			}
            		zgeqpx_ ( &JOB,&M,&N,&K,Ac,&LDA,Cc,&LDC,JPVT,&IRCOND, &ORCONDr,&RANK,SVLUESr,WORKc,&LWORK,RWORK,&INFO ); 

			plhs[0]  = mxCreateDoubleMatrix(LDA, N, 1);
			RE = mxGetPr(plhs[0]); 
			IM = mxGetPi(plhs[0]); 
			for ( i = 0; i < LDA*N; i++) {
				RE[i] = creal(Ac[i]); 
				IM[i] = cimag(Ac[i]);  
			}


 			plhs[1] = mxCreateDoubleMatrix(N, 1, 0); 
			RE = mxGetPr(plhs[1]); 
			for ( i = 0; i<N; i++) {
				RE[i] = (double) JPVT[i]; 
			}

			plhs[2] = mxCreateDoubleMatrix(1, 1, 0); 
			*mxGetPr(plhs[2])=(double) RANK; 

			free(Ac); 
			free(Cc); 
			free(WORKc); 
			free(RWORK); 


		} else { 
			double *RE; 
			Ar = (double *) malloc(sizeof(double) * LDA * N); 
			Cr = (double *) malloc(sizeof(double) * LDA * N); 
			WORKr = (double *) malloc(sizeof(double) * LWORK); 

			memcpy(Ar, mxGetPr(prhs[0]), sizeof(double)*LDA*N); 
            		dgeqpx_( &JOB, &M, &N, &K, Ar, &LDA, Cr, &LDC, JPVT, &IRCOND,& ORCONDr, &RANK, SVLUESr, WORKr, &LWORK, &INFO ); 
			plhs[0] = mxCreateDoubleMatrix(LDA, N, 0); 
			memcpy(mxGetPr(plhs[0]),Ar, sizeof(double)*LDA*N); 

 			plhs[1] = mxCreateDoubleMatrix(N, 1, 0); 
			RE = mxGetPr(plhs[1]); 
			for ( i = 0; i<N; i++) {
				RE[i] = (double) JPVT[i]; 
			}

			plhs[2] = mxCreateDoubleMatrix(1, 1, 0); 
			*mxGetPr(plhs[2])=(double) RANK; 
			free(Ar); 
			free(Cr); 
			free(WORKr); 

	    	}
		free(JPVT); 
	} 
	/* [Q,R,p,r] = rrqr ... */ 
      	else if(nlhs == 4) {
		if (nrhs >= 4 ||  nrhs <= 1) {
			mexErrMsgTxt("Two or three input arguments required."); 
		} else if ( nrhs == 3) {
			double *tmp = mxGetPr(prhs[2]); 
			IRCOND = *tmp; 
		        if(IRCOND < 0 ){ 
				       mexErrMsgTxt("tol must be greater then zero."); 
		        }
       		} else { 
			IRCOND = 0;  /*  !maps to machine precision internally */
       		}

		JOB = 3;
		M = mxGetM(prhs[0]);
		N = mxGetN(prhs[0]);
		MC = mxGetM(prhs[1]); 
		NC = mxGetN(prhs[1]);
		K = MC;
		LDA = M;
		LDC = K;
		MN = min( M, N );
		LWORK = 2*MN+2*N+max(K,N);

		JPVT = (F77_INT *) malloc ( sizeof(F77_INT) *N); 
            	ic = mxIsComplex(prhs[0]); 
#ifdef DEBUG		
		printf("JOB(1)=%d\n",JOB); 
		printf("M  (2)=%d\n",M); 
		printf("N  (3)=%d\n",N); 
		printf("K  (4)=%d\n",K); 
		printf("LDA(6)=%d\n",LDA); 
#endif 

		if (ic) {
			int i;
			double *RE; 
			double *IM; 
			Ac = (double complex *) malloc(sizeof(double complex) * LDA *N); 
			Cc = (double complex *) malloc(sizeof(double complex) *LDA *N); 
			WORKc = (double complex *) malloc( sizeof (double complex) *LWORK); 
            		RW = 2*N; 
			RWORK = (double *) malloc(sizeof(double)* RW); 


			RE = mxGetPr(prhs[0]); 
			IM = mxGetPi(prhs[0]); 
			for ( i = 0; i < LDA*N; i++) {
				Ac[i] = RE[i] + IM[i]*I; 
			}
			RE = mxGetPr(prhs[1]); 
			IM = mxGetPi(prhs[1]); 
			for ( i = 0; i < MC*NC; i++) {
				Cc[i] = RE[i] + IM[i]*I; 
			}
            	
			zgeqpx_(&JOB, &M, &N, &K, Ac, &LDA, Cc, &LDC, JPVT, &IRCOND,&ORCONDr,&RANK,SVLUESr,WORKc,&LWORK,RWORK,&INFO); 
			plhs[0] = mxCreateDoubleMatrix(MC, NC, 1); 
			RE = mxGetPr(plhs[0]); 
			IM = mxGetPi(plhs[0]); 
			for ( i = 0; i < MC*NC; i++) {
				RE[i] = creal(Cc[i]); 
				IM[i] = cimag(Cc[i]);  
			}

			plhs[1]  = mxCreateDoubleMatrix(LDA, N, 1);
			RE = mxGetPr(plhs[1]); 
			IM = mxGetPi(plhs[1]); 
			for ( i = 0; i < LDA*N ; i++) {
				RE[i] = creal(Ac[i]); 
				IM[i] = cimag(Ac[i]);  
			}
	
			plhs[2] = mxCreateDoubleMatrix(N, 1, 0); 
			RE = mxGetPr(plhs[2]); 
			for ( i = 0; i<N; i++) {
				RE[i] = (double) JPVT[i]; 
			}

			plhs[3] = mxCreateDoubleMatrix(1, 1, 0); 
			*mxGetPr(plhs[3])=(double) RANK; 

            		free(Ac); 
			free(Cc); 
			free(WORKc); 
			free(RWORK); 
		} else {
			double *RE; 
			Ar = (double *) malloc(sizeof(double) * LDA * N); 
			Cr = (double *) malloc(sizeof(double) * MC * NC); 
			WORKr = (double *) malloc(sizeof(double) * LWORK); 
			memcpy(Ar, mxGetPr(prhs[0]), sizeof(double)*LDA*N); 
			memcpy(Cr, mxGetPr(prhs[1]), sizeof(double)*MC*NC); 
            		
			dgeqpx_( &JOB, &M, &N, &K, Ar, &LDA, Cr, &LDC, JPVT, &IRCOND, &ORCONDr, &RANK, SVLUESr, WORKr, &LWORK, &INFO ); 
			plhs[0] = mxCreateDoubleMatrix(MC,NC, 0); 
			memcpy(mxGetPr(plhs[0]),Cr, sizeof(double)*MC*NC); 
			plhs[1] = mxCreateDoubleMatrix(LDA,N, 0); 
			memcpy(mxGetPr(plhs[1]),Ar, sizeof(double)*LDA*N); 

 			plhs[2] = mxCreateDoubleMatrix(N, 1, 0); 
			RE = mxGetPr(plhs[2]); 
			for ( i = 0; i<N; i++) {
				RE[i] = (double) JPVT[i]; 
			}

			plhs[3] = mxCreateDoubleMatrix(1, 1, 0); 
			*mxGetPr(plhs[3])=(double) RANK; 
			free(Ar); 
			free(Cr); 
			free(WORKr); 
		}
		free(JPVT); 
	}  else { 
		mexErrMsgTxt("Three or four output arguments required."); 
	}
	return ; 
}


